#include "main.h"
#ifdef AVR
#include <avr/pgmspace.h>
#define rd(i) pgm_read_byte_near(&i)
#else
#define rd(i) (i)
#endif

#define HEADSIZE 90
#define BLKSIZE 256

typedef struct state {
	uint8_t	 predictor;
	uint16_t delta;
	int16_t	 sample1, sample2;
} state;

/*
	Compute a linear PCM value from the given differential coded
	value.
*/
static inline int16_t decode_sample (struct state *state,
	uint8_t code, const int16_t *coefficient) {
	const int32_t adaptive[] = {
		230, 230, 230, 230, 307, 409, 512, 614,
		768, 614, 512, 409, 307, 230, 230, 230
	};
	int32_t	linearSample, delta;

	linearSample = ((state->sample1 * coefficient[0]) +
		(state->sample2 * coefficient[1])) / 256;
	printf("ls=%d c0=%d c1=%d code=%d\n", linearSample, coefficient[0], coefficient[1], code);
	if (code & 0x08)
		linearSample += state->delta * (code-0x10);
	else
		linearSample += state->delta * code;

	/* Clamp linearSample to a signed 16-bit value. */
#if 0
	if (linearSample < INT16_MIN)
		linearSample = INT16_MIN;
	else if (linearSample > INT16_MAX)
		linearSample = INT16_MAX;
#endif

	delta = ((int32_t) state->delta * adaptive[code])/256;
	if (delta < 16)
	{
		delta = 16;
	}

	state->delta = delta;
	state->sample2 = state->sample1;
	state->sample1 = linearSample;

	/* Because of earlier range checking, new_sample will be
	   in the range of an int16_t. */
	return (int16_t) linearSample;
}

static const int16_t coeff[] = { 256, 0, 512, -256,
			  0, 0, 192, 64, 240,
			  0, 460, -208, 392, -232 };

/* Decode one block of MS ADPCM data. */
static void decode_block (const uint8_t *in) {
	int i;
	const int16_t *c;
	state state;

	/* Initialize predictor. */
	state.predictor = rd(*in++);
	assert(state.predictor < 7);

	/* Initialize delta. */
	state.delta = (rd(in[1])<<8) | rd(in[0]);
	in += sizeof (uint16_t);

	/* Initialize first two samples. */
	state.sample1 = (rd(in[1])<<8) | rd(in[0]);
	in += sizeof (uint16_t);

	state.sample2 = (rd(in[1])<<8) | rd(in[0]);
	in += sizeof (uint16_t);

	c = &coeff[state.predictor*2];

	write_dac(state.sample2);
	write_dac(state.sample1);

	/* The first two samples have already been 'decoded' in
	   the block header. */
	for (i = 498; i>0; i-=2, in++) {
		uint8_t b = rd(*in);
		printf("%d s1=%d s2=%d pr=%d de=%d b=%d\n", i, state.sample1, state.sample2, state.predictor, state.delta, b);
		write_dac(decode_sample(&state, b>>4, c));
		write_dac(decode_sample(&state, b&15, c));
	}
}

void play_adpcm (const uint8_t *data, const uint8_t *end) {
	//printf("INT16_MIN=%d, INT16_MAX=%d\n", INT16_MIN, INT16_MAX);
	data += HEADSIZE;
	//int i=0;
	do {
		//fprintf(stderr, "%d %x %x\n", i++, data, end);
		decode_block (data);
		data += BLKSIZE;
	} while (data+BLKSIZE <= end);
}
